import sys, os, sets, math, random

import pygame
pygame.init()
from pygame.locals import *

from pgu import tilevid, timer, ani

CODE_START = 1
CODE_LADDER = 3

TILE_CLEAR = 0
TILE_LADDER = 2
TILE_LADDER_PLATFORM = 3
CLEAR_TILES = sets.Set([TILE_CLEAR] + range(48, 256))
l = []
for i in range(8):
    l.extend([56+i*16, 57+i*16])
HITTABLE_TILES = sets.Set(l)
TILE_RAILS = 42

DIRECTION_UP=0
DIRECTION_RIGHT=1
DIRECTION_DOWN=2
DIRECTION_LEFT=3

spanner_hit_offsets = {
    DIRECTION_UP:    (4, -24),
    DIRECTION_RIGHT: (4, -4),
    DIRECTION_DOWN:  (-10, -4),
    DIRECTION_LEFT:  (-24, -4), 
}
spanner_hit_point_offsets = {
    DIRECTION_UP:    (-8, -36),
    DIRECTION_RIGHT: (4, -16),
    DIRECTION_DOWN:  (-24, 0),
    DIRECTION_LEFT:  (-40, -16), 
}
FPS = 25

# these affect the difficulty
MAX_PLAYER_HEALTH = 10
NOZZLE_SPEED = 3         # pixels per frame
NOZZLE_RECHARGE = 15     # frames
TIME_ALLOWED = 40.       # seconds
MAX_STEAM_LEVEL = 100.
#MAX_STEAM_LEVEL = 10.

modenine = pygame.font.Font('data/sotc/MODENINE.TTF', 12)
modenine_small = pygame.font.Font('data/sotc/MODENINE.TTF', 8)

player_won = pygame.image.load('data/sotc/win.png')

# places to put explosions
EXPLOSION_PLACES = []
for x in (86, 536):
    for y in (44, 140, 236, 332):
        EXPLOSION_PLACES.append((x, y))
LARGE_EXPLOSION_PLACES = []
for y in (2, 3, 4, 5, 6, 7, 8):
    for x in (8, 9, 10, 11):
        LARGE_EXPLOSION_PLACES.append((x*32-16, y*32-16))

# load all the sounds
sounds = {}
for file in os.listdir('data/sotc'):
    if not file.endswith('.ogg'): continue
    name = os.path.splitext(file)[0]
    if pygame.mixer.get_init():
        sounds[name] = pygame.mixer.Sound(os.path.join('data/sotc', file))
    else:
        sounds[name] = None

class Game(tilevid.Tilevid):
    def __init__(self, screen):
        tilevid.Tilevid.__init__(self)
        self.screen = screen
        self.screen.fill((0,0,0))

        cdata = {
            #TILE_CLEAR: ('player', self.player_fall, {}),
        }
        for tileid in range(16, 48):    # 1st to 3rd rows
            cdata[tileid] = ('player', self.tile_block, {})

        self.tga_load_tiles('data/sotc/level1-tiles.tga', (32, 32), cdata)
        self.tga_load_level('data/sotc/level1.tga', 1)
        self.bounds = pygame.Rect(32, 32, (len(self.tlayer[0])-2)*32,
            (len(self.tlayer)-2)*32)

        # animations are:
        # up    (l, m, r)
        # right (l, m, r)
        # down  (l, m, r)
        # left  (l, m, r)
        # (as per DIRECTION_ number ordering above)
        ani.ani_load(self, 'player', pygame.image.load('data/sotc/player.tga'),
            (32, 32), (8, 8, 16, 24), [4, 3])

        # pics are for up, left, down, right (order as above)
        ani.ani_load(self, 'spanner',
            pygame.image.load('data/sotc/spanner-held.tga'),
            (32, 32), (0, 0, 32, 32), [4])

        # hit pics are aligned top-left in the sprite, and ordered as above
        ani.ani_load(self, 'spanner-hit',
            pygame.image.load('data/sotc/spanner-hit.tga'),
            (32, 32), (0, 0, 32, 32), [4])

        # sparks are centered at 16, 13
        ani.ani_load(self, 'hit-sparks',
            pygame.image.load('data/sotc/hit-sparks.tga'),
            (32, 32), (0, 0, 32, 32), [6])

        # figure player start
        for y, row in enumerate(self.clayer):
            for x, cell in enumerate(row):
                if cell == CODE_START:
                    player_pos = (x, y)
        pos = self.tile_to_screen(player_pos)
        self.set_player_sprite(DIRECTION_DOWN, 1,
            (pos[0]+8, pos[1]+8, 16, 24), MAX_PLAYER_HEALTH)
        self.spanner= None
        self.hit_spanner= None
        self.hit_sparks = []

        # steam nozzles
        ani.ani_load(self, 'nozzle', pygame.image.load('data/sotc/nozzle.tga'),
            (32, 32), (0, 0, 32, 32), [2])
        ani.ani_load(self, 'steam-jet-left',
            pygame.image.load('data/sotc/steam-jet-left.tga'),
            (160, 32), (0, 0, 160, 32), [4])
        ani.ani_load(self, 'steam-jet-right',
            pygame.image.load('data/sotc/steam-jet-right.tga'),
            (160, 32), (0, 0, 160, 32), [4])

        # figure nozzle start
        self.nozzles = []
        s = tilevid.Sprite(self.images['nozzle.0'], (192-8, 0, 32, 32))
        self.sprites.append(s)
        self.nozzles.append(s)
        s.loop = self.nozzle_loop
        s.direction = DIRECTION_LEFT
        s.move = DIRECTION_DOWN
        s.steam = None
        s.recharge = 0

        s = tilevid.Sprite(self.images['nozzle.1'], (416+8, 320, 32, 32))
        self.sprites.append(s)
        self.nozzles.append(s)
        s.loop = self.nozzle_loop
        s.direction = DIRECTION_RIGHT
        s.move = DIRECTION_UP
        s.steam = None
        s.recharge = 0

        # boiler fires
        # XXX sprite ordering bites me inna ass here
#        ani.ani_load(self, 'boiler-fire',
#            pygame.image.load('data/sotc/boiler-fire.png'),
#            (16, 16), (0, 0, 16, 16), [4])
#        for x in (86, 536):
#            for y in (44, 140, 236, 332):
#                frame = random.randint(0,3)
#                s = tilevid.Sprite(self.images['boiler-fire.%s'%frame],
#                    (x, y, 16, 16))
#                self.sprites.append(s)
#                s.frame = frame
#                s.anim = random.randint(1, 5)
#                s.loop = self.boiler_fire_loop

        # colossus
        self.steam_level = MAX_STEAM_LEVEL

        ani.ani_load(self, 'colossus-eyes',
            pygame.image.load('data/sotc/colossus-eyes.png'),
            (16, 16), (0, 0, 16, 16), [6])
        s = tilevid.Sprite(self.images['colossus-eyes.0'], (301, 108, 16, 16))
        self.sprites.append(s)
        s.loop = self.colossus_eyes_loop
        s.anim = 5
        s.frame = 0
        s.anim = 4
        s.direction = 1
        s = tilevid.Sprite(self.images['colossus-eyes.0'], (325, 108, 16, 16))
        self.sprites.append(s)
        s.loop = self.colossus_eyes_loop
        s.anim = 5
        s.frame = 0
        s.anim = 4
        s.direction = 1

        # boom boom
        ani.ani_load(self, 'explosion',
            pygame.image.load('data/sotc/explosion.png'), (40, 40),
            (0,0,40,40), [14])
        ani.ani_load(self, 'large-explosion',
            pygame.image.load('data/sotc/large_explosion.png'), (48, 48),
            (0,0,48,48), [8])

        # game-state vars
        self.quit = 0
        self.died = 0
        self.won = 0
        self.time_left = TIME_ALLOWED * FPS

    def set_player_sprite(self, direction, frame, rect, health=None):
        if health is not None:
            s = tilevid.Sprite(self.images['player.%s.%s'%(direction,
                frame)], rect)
            self.sprites.append(s)
            s.loop = self.player_loop
            s.groups = self.string2groups('player')
            s.health = health
            self.player = s
        else:
            s = self.player
            s.image = self.images['player.%s.%s'%(direction, frame)][0]
        s.direction = direction
        s.animation_time = 3            # frames to new anim frame
        s.animation_frame = frame       # frame to display

    def boiler_fire_loop(self, game, fire):
        fire.anim -= 1
        if fire.anim:
            return
        fire.frame = (fire.frame + 1)%4
        fire.anim = random.randint(2,4)
        fire.image = self.images['boiler-fire.%s'%fire.frame][0]

    def colossus_eyes_loop(self, game, eye):
        eye.anim -= 1
        if eye.anim:
            return
        eye.anim = 4
        if self.won and eye.direction != 1:
            eye.direction = 1
        eye.frame += eye.direction
        if eye.frame == -1:
            eye.frame = 0
            eye.direction = 1
        if eye.frame == 6:
            eye.frame = 5
            if self.won:
                eye.direction = -1
        eye.image = self.images['colossus-eyes.%s'%eye.frame][0]

    def player_loop(self, game, player):
        keys = pygame.key.get_pressed()

        cx, cy = player.rect.center
        tx, ty = self.screen_to_tile((cx, cy))
        tile = self.tlayer[ty][tx]
        direction = player.direction

        dx,dy = 0,0
        maybe_direction = None
        new_direction = None
        anim = False
        hold_spanner = False
        hit_spanner = False
        if keys[K_PRINT]:
            pygame.image.save(self.screen, '/tmp/sotc-screenshot.bmp')
        if not self.won or self.died:
            if keys[K_LSHIFT] or keys[K_RSHIFT]:
                if keys[K_LEFT]:
                    hit_spanner = True
                    maybe_direction = DIRECTION_LEFT
                elif keys[K_RIGHT]:
                    hit_spanner = True
                    maybe_direction = DIRECTION_RIGHT
                elif keys[K_UP]:
                    hit_spanner = True
                    maybe_direction = DIRECTION_UP
                elif keys[K_DOWN]:
                    hit_spanner = True
                    maybe_direction = DIRECTION_DOWN
                else:
                    hold_spanner = True
            else:
                if keys[K_LEFT]:
                    anim = True
                    dx-=1
                    if direction == DIRECTION_LEFT:
                        new_direction = DIRECTION_LEFT
                    else: maybe_direction = DIRECTION_LEFT
                if keys[K_RIGHT]:
                    anim = True
                    dx+=1
                    if direction == DIRECTION_RIGHT:
                        new_direction = DIRECTION_RIGHT
                    else: maybe_direction = DIRECTION_RIGHT
                if tile in (TILE_LADDER, TILE_LADDER_PLATFORM):
                    if keys[K_UP]:
                        anim = True
                        dy-=1
                        if direction == DIRECTION_UP:
                            new_direction = DIRECTION_UP
                        else: maybe_direction = DIRECTION_UP
                    if keys[K_DOWN]:
                        anim = True
                        dy+=1
                        if direction == DIRECTION_UP:
                            new_direction = DIRECTION_UP
                        else: maybe_direction = DIRECTION_UP
        if tile in CLEAR_TILES:
            dy += 1

        rect = player.rect
        if new_direction is None and maybe_direction is not None:
            direction = new_direction = maybe_direction
            if hold_spanner or hit_spanner:
                self.set_player_sprite(maybe_direction, 1, rect)
            else:
                self.set_player_sprite(maybe_direction, 0, rect)
        elif anim:
            player.animation_time -= 1
            if not player.animation_time:
                frame = (player.animation_frame + 1)%3
                self.set_player_sprite(direction, frame, rect)
        elif player.animation_frame != 1:
            self.set_player_sprite(direction, 1, rect)

        player.rect.x += dx*4
        player.rect.y += dy*4
        player.rect.clamp_ip(game.view)

        # update hit sparks
        l = []
        for spark in self.hit_sparks:
            frame = spark.animation_frame + 1
            if frame == 6:
                self.sprites.remove(spark)
                continue
            spark.image = self.images['hit-sparks.%s'%frame][0]
            spark.animation_frame = frame
            l.append(spark)
        self.hit_sparks = l

        if hit_spanner and not self.hit_spanner:
            # render the hit spanner in the correct direction
            cx, cy = player.rect.center
            ox, oy = spanner_hit_offsets[direction]
            s = tilevid.Sprite(self.images['spanner-hit.%s'%direction],
                (cx+ox, cy+oy, 32, 32))
            self.sprites.append(s)
            s.direction = new_direction
            self.hit_spanner = s
            self.hit_spanner.anim = 4

            ox, oy = spanner_hit_point_offsets[direction]
            htx, hty = self.screen_to_tile((cx+ox+16, cy+oy+16))
            tile = self.tlayer[hty][htx]
            if tile in HITTABLE_TILES:
                self.steam_level -= 2
                if self.steam_level < 0:
                    self.steam_level = 0
                    self.player_won()
                # detect damage done
                s = tilevid.Sprite(self.images['hit-sparks.0'],
                    (cx+ox, cy+oy, 32, 32))
                s.animation_frame = 0
                self.sprites.append(s)
                self.hit_sparks.append(s)
                self.tlayer[hty][htx] = tile + 32
                sounds['hammer%d'%random.randint(1,4)].play()

        elif self.hit_spanner:
            self.hit_spanner.anim -= 1
            if not self.hit_spanner.anim:
                self.sprites.remove(self.hit_spanner)
                self.hit_spanner = None

        if hold_spanner:
            if not self.spanner:
                s = tilevid.Sprite(self.images['spanner.%s'%direction],
                    player.rect)
                self.sprites.append(s)
                s.direction = new_direction
                self.spanner = s
            px, py = player.rect.center
            self.spanner.rect.center = (px, py-8)
        elif self.spanner:
            self.sprites.remove(self.spanner)
            self.spanner = None

    def nozzle_loop(self, game, nozzle):
        if nozzle.move == DIRECTION_DOWN:
            nozzle.rect.y += NOZZLE_SPEED
            tx,ty = self.screen_to_tile(nozzle.rect.midbottom)
            tile = self.tlayer[ty][tx]
            if tile != TILE_RAILS:
                nozzle.move = DIRECTION_UP
                nozzle.rect.y -= NOZZLE_SPEED
        else:
            nozzle.rect.y -= NOZZLE_SPEED
            tx,ty = self.screen_to_tile(nozzle.rect.midtop)
            tile = self.tlayer[ty][tx]
            if tile != TILE_RAILS:
                nozzle.move = DIRECTION_DOWN
                nozzle.rect.y += NOZZLE_SPEED

        # if recharging DO NOTHING ELSE
        if nozzle.recharge:
            nozzle.recharge -= 1
            return

        # see if we should shoot
        nx, ny = nozzle.rect.midtop
        if nozzle.direction == DIRECTION_LEFT:
            rect = pygame.Rect(nx - 160, ny, 160, 32)
            jet = 'left'
        if nozzle.direction == DIRECTION_RIGHT:
            rect = pygame.Rect(nx, ny, 160, 32)
            jet = 'right'
        if self.player.rect.colliderect(rect):
            s = tilevid.Sprite(self.images['steam-jet-%s.%s'%(jet, 0)], rect)
            self.sprites.append(s)
            s.jet = jet
            s.frame = 0
            s.anim = 1
            s.loop = self.steam_loop
            nozzle.recharge = NOZZLE_RECHARGE
            sounds['steam1'].play()

    def dead_nozzle_loop(self, game, nozzle):
        nozzle.rect.y += NOZZLE_SPEED*2
        tx,ty = self.screen_to_tile(nozzle.rect.midbottom)
        tile = self.tlayer[ty][tx]
        if tile != TILE_RAILS:
            nozzle.rect.y -= NOZZLE_SPEED*2

    def tile_block(self, game, t, a):
        if (a._rect.bottom <= t._rect.top and a.rect.bottom > t.rect.top):
            a.rect.bottom = t.rect.top
        if (a._rect.right <= t._rect.left and a.rect.right > t.rect.left):
            a.rect.right = t.rect.left
        if (a._rect.left >= t._rect.right and a.rect.left < t.rect.right):
            a.rect.left = t.rect.right
        if (a._rect.top >= t._rect.bottom and a.rect.top < t.rect.bottom):
            a.rect.top = t.rect.bottom


    def steam_loop(self, game, steam):
        # animate that steam
        frame = steam.frame + 1
        if frame > 3:
            self.sprites.remove(steam)
        elif steam.anim:
            steam.anim -= 1
        else:
            steam.image = self.images['steam-jet-%s.%s'%(steam.jet, frame)][0]
            steam.frame = frame
            steam.anim = frame
            if self.player.rect.colliderect(steam.rect):
                self.damage_player(1)

    def update(self, s):
        l  = []

        if self.won:
            self.paint(s)
            l.append(s.blit(player_won, (0,0)))

        elif self.died:
            self.paint(s)
            l.append(s.blit(player_died, (0,0)))

        else:
            # timers affect these tiles
            for y in (5, 6, 7):
                for x in (9, 10):
                    l.append(s.blit(self.tiles[self.tlayer[y][x]].image,
                        (x*32,y*32)))

            # timer bar
            time = (60.-self.time_left/FPS)/60.
            angle = (90 + 360.*time) % 360
            angle = math.radians(angle)
            sx, sy = (319, 229)
            ex = math.cos(angle) * 16 + 319
            ey = 229 - math.sin(angle) * 16
            l.append(pygame.draw.line(s, (0,0,0), (sx, sy), (ex, ey)))

            # steam level
            sl = self.steam_level + random.choice((-2, -1, 0, 0, 0, 0, 1, 2))
            angle = (90 + (360. * (MAX_STEAM_LEVEL - sl)/MAX_STEAM_LEVEL)) % 360
            angle = math.radians(angle)
            sx, sy = (319, 184)
            ex = math.cos(angle) * 16 + 319
            ey = 184 - math.sin(angle) * 16
            l.append(pygame.draw.line(s, (0,0,0), (sx, sy), (ex, ey)))

            l.extend(tilevid.Tilevid.update(self, s))

            # player health
            r = s.get_rect()
            hbar = pygame.Rect(0, 0, 116, 24)
            hbar.midbottom = r.midbottom
            l.append(hbar)
            s.fill((0,0,0), hbar)
            cur_health = pygame.Rect(hbar)
            cur_health.height -= 8
            cur_health.y += 4
            cur_health.x += 8
            cur_health.width = (self.player.health * 100. / MAX_PLAYER_HEALTH)
            s.fill((255,0,0), cur_health)

        if '-showfps' in sys.argv:
            fps = int(self.timer.get_fps())
            text = modenine.render('FPS: %s'%fps, 1, (200, 200, 200), (0, 0, 0))
            tx, ty = text.get_size()
            sx, sy = s.get_size()
            rect = (0, sy-ty, 100, ty)
            s.fill((0, 0, 0), rect)
            s.blit(text, (0, sy-ty))
            l.append(rect)

        return l

    def damage_player(self, amount):
        self.player.health = max(0, self.player.health - amount)
        if not self.player.health:
            self.died = 1

    def explosion_loop(self, game, explosion):
        explosion.frame += 1
        if explosion.frame < 14:
            explosion.image = self.images['explosion.%s'%explosion.frame][0]
        else:
            explosion.image = self.images['explosion.0'][0]
            explosion.rect.topleft = random.choice(EXPLOSION_PLACES)
            explosion.frame = 0
            sounds['explosion%d'%random.randint(1,3)].play()

    def large_explosion_loop(self, game, explosion):
        explosion.anim -= 1
        if explosion.anim:
            return

        explosion.anim = 2
        explosion.frame += 1
        if explosion.frame < 8:
            explosion.image = self.images['large-explosion.%s'%
                explosion.frame][0]
        else:
            explosion.image = self.images['large-explosion.0'][0]
            explosion.rect.topleft = random.choice(LARGE_EXPLOSION_PLACES)
            explosion.frame = 0
            sounds['explosion%d'%random.randint(1,3)].play()

    def player_won(self):
        self.won = 1

        for nozzle in self.nozzles:
            nozzle.loop = self.dead_nozzle_loop

        pos = random.choice(EXPLOSION_PLACES)
        s = tilevid.Sprite(self.images['explosion.0'], pos + (40, 40))
        s.frame = 0
        self.sprites.append(s)
        s.loop = self.explosion_loop

        pos = random.choice(LARGE_EXPLOSION_PLACES)
        s = tilevid.Sprite(self.images['large-explosion.0'], pos + (48, 48))
        s.anim = 2
        s.frame = 0
        self.sprites.append(s)
        s.loop = self.large_explosion_loop

    def run(self): 
        self.paint(self.screen)
        pygame.display.flip()

        self.timer = pygame.time.Clock()
        while not self.quit:
            for e in pygame.event.get():
                if e.type is QUIT: self.quit = 1
                if e.type is KEYDOWN:
                    if e.key == K_ESCAPE: self.quit = 1
                    if self.won and e.key == K_SPACE: raise PlayerWon()

            self.loop()

            if not self.won and not self.died:
                self.time_left -= 1

            if not self.time_left or self.died:
                raise PlayerLost()
            pygame.display.update(self.update(self.screen))
            self.timer.tick(FPS)

class PlayerLost(Exception):
    pass
class PlayerWon(Exception):
    pass
class ExitGame(Exception):
    pass

class Title:
    def __init__(self, screen, image):
        title = pygame.image.load('data/sotc/%s'%image)
        screen.blit(title, (0,0))
        pygame.display.flip()

    def run(self):
        self.timer = pygame.time.Clock()
        while 1:
            for e in pygame.event.get():
                if e.type is KEYDOWN and e.key == K_ESCAPE: raise ExitGame
                if e.type is KEYDOWN and e.key == K_SPACE: return
            self.timer.tick(5)
        

if '-fs' in sys.argv:
    screen = pygame.display.set_mode((640, 480), FULLSCREEN)
else:
    screen = pygame.display.set_mode((640, 480), SWSURFACE)
pygame.display.set_caption('Steam of the Colossus!')
while 1:
    try:
        Title(screen, 'title-screen.png').run()
        try:
            Game(screen).run()
        except PlayerLost:
            Title(screen, 'lose.png').run()
        except PlayerWon:
            pass
            #Title(screen, 'win.png').run()
    except ExitGame:
        break

